#include <fstream>
#include <algorithm>
#include "LogicProgramProblem.hpp"

#include "parser/Driver.h"

using namespace std;
using namespace lpopt;

LogicProgramProblem::LogicProgramProblem(istream *stream, AbstractHypertreeDecompositionAlgorithm *htdalg, bool benchmark, bool ignoreHeadVariables)
	: Problem(htdalg, benchmark), benchmark(benchmark), ignoreHeadVariables(ignoreHeadVariables)
{
	this->parsingDriver = new parser::Driver(stream);
	program = NULL;
	createdProgram = NULL;
	gen = NULL;
	width = 0;
}

LogicProgramProblem::~LogicProgramProblem()
{
	delete parsingDriver;
	delete program;
	if (createdProgram != NULL)
	  delete new Program(createdProgram);
}

#pragma GCC diagnostic ignored "-Wmaybe-uninitialized" // beg
void LogicProgramProblem::parse()
{
	clock_t beg;
	if (benchmark) beg = clock();
	delete program;
	program = this->parsingDriver->parse();
	if (benchmark)
	{
	  clock_t thisTime = clock() - beg;
	  parse_time += thisTime;
	  cout << "parsing finished. took " << (float) thisTime / CLOCKS_PER_SEC << " seconds." << endl;
	}
}

#pragma GCC diagnostic ignored "-Wmaybe-uninitialized" // beg
void LogicProgramProblem::preprocess()
{
	clock_t beg;
	if (benchmark) beg = clock();
	program->splitPooling();
	StringGenerator innerGen(new string("temp_innerDomain_"));
	program->optimizeInnerDomains(innerGen);
	delete new Program(createdProgram);
	createdProgram = new vector<Statement *>();
	if (benchmark)
	{
	  clock_t thisTime = clock() - beg;
	  me_time += thisTime;
	  cout << "\t\t" << (float) thisTime / CLOCKS_PER_SEC << " seconds for inner domain optimization and pooling splitting. (finished)" << endl;
	}
}

#pragma GCC diagnostic ignored "-Wmaybe-uninitialized" // beg
Solution *LogicProgramProblem::calculateSolution(AbstractHTDAlgorithm *algorithm) // iterates over all statements and calculates a decomp & solution for each
{
	this->currentStatement = 0;
	if (benchmark)
	{
	  htdlib_time = 0;
	  parse_time = 0;
	  me_time = 0;
	}
	do {
	  string* tmpString = new string("temp_decomp_");
	  tmpString->append(std::to_string(currentStatement)).append("_");
	  gen = new StringGenerator(tmpString);
	  clock_t old_parse_time, old_me_time, beg;
	  if (benchmark)
	  {
	    old_parse_time = parse_time;
	    old_me_time = me_time;
	    beg = clock();
	  }
	  ExtendedHypertree *htd = this->calculateHypertreeDecomposition(); // [parse(), preprocess(),][if not already parsed] buildHypergraphRepresentation(), a tree decomposition is generated, prepareHypertreeDecomposition()
	  if (benchmark)
	  {
	    htdlib_time += clock() - beg - (me_time - old_me_time) - (parse_time - old_parse_time);
	  }
	  if (benchmark) beg = clock();
	  this->calculateSolutionFromDecomposition(algorithm, htd);
	  if (benchmark) me_time += clock() - beg;
	  delete gen;
	  delete htd;
	  ++currentStatement;
	} while(currentStatement < program->statementsSize());
	if (benchmark)
	{
	  cout << "done optimizing logic program. took " << (float) (htdlib_time + me_time) / CLOCKS_PER_SEC << " seconds." << endl;
	  cout << "\t" << (float) me_time / CLOCKS_PER_SEC << " seconds for inner domain optimization, pooling splitting, hypergraph representation building and final solution calculation based on decomp," << endl;
	  cout << "\t" << (float) htdlib_time / CLOCKS_PER_SEC << " seconds for tree decomposing library." << endl;
	}
	return NULL;
}

#pragma GCC diagnostic ignored "-Wmaybe-uninitialized" // beg
ExtendedHypertree *LogicProgramProblem::prepareHypertreeDecomposition(ExtendedHypertree *node) // determines head node of the given decomposition
{
	if (currentStatement == 0 && program->statementsSize() == 0) // just parsed. check if we have more than 0 statements.
	  return node;

	if(ignoreHeadVariables) {
		node->setRoot();
		return node; // TODO maybe select bag with most head-included variables
	}

	clock_t beg;
	if (benchmark) beg = clock();

	const VertexSet &vertices = node->getVertices();
	
	varset* headVars = curStatement().headVars();
	
	VertexSet headVertices;
	for (varset::iterator it = headVars->begin();  it != headVars->end(); ++it)
		headVertices.insert(this->getVertexId((*it)->getName()));
	delete headVars;
	
	if(std::includes(vertices.begin(),  vertices.end(),  headVertices.begin(),  headVertices.end()))
	{
		node->setRoot();
		if (benchmark) me_time += clock() - beg;
		return node;
	}

	for(list<Hypertree *>::iterator it = node->getChildren()->begin(); it != node->getChildren()->end(); ++it)
	{
		if (benchmark) me_time += clock() - beg;
		ExtendedHypertree *r = this->prepareHypertreeDecomposition((ExtendedHypertree *)*it);
		if (benchmark) beg = clock();
		if(r != NULL) {
		  if (benchmark) me_time += clock() - beg;
		  return r;
		}
	}
	if (benchmark) me_time += clock() - beg;
	return NULL;
}

const Statement& LogicProgramProblem::curStatement() const
{
	return program->getStatement(currentStatement);
}

#pragma GCC diagnostic ignored "-Wmaybe-uninitialized" // beg
Hypergraph *LogicProgramProblem::buildHypergraphRepresentation() // builds representation only for current statement
{
	clock_t beg;
	if (benchmark) beg = clock();

	VertexSet vertices;
	HyperedgeSet edges;

	if (currentStatement == 0 && program->statementsSize() == 0) // just parsed. check if we have more than 0 statements.
	{
	  if (benchmark) 
	  {
	    clock_t thisTime = clock() - beg;
	    me_time += thisTime;
	  }
	  return createHypergraphFromSets(vertices, edges);
	}

	this->clearVertexStore();
	
	// vertices:
	varset* globalVars = curStatement().variables();
	for (varset::iterator it = globalVars->begin(); it != globalVars->end(); ++it)
		vertices.insert(storeVertexName((*it)->getName()));
	delete globalVars;
	
	// edges:
	set<varset*>* elementVars = ignoreHeadVariables ? curStatement().bodyVars() : curStatement().elementVars();
	for (set<varset*>::iterator it1 = elementVars->begin(); it1 != elementVars->end(); ++it1)
	{
		VertexSet vs;
		for (varset::iterator it2 = (*it1)->begin(); it2 != (*it1)->end(); ++it2)
			vs.insert(getVertexId((*it2)->getName()));
		edges.insert(vs);
		delete *it1;
	}
	delete elementVars;

	if (benchmark) me_time += clock() - beg;
	return createHypergraphFromSets(vertices, edges);
}

Program* LogicProgramProblem::getOptimizedProgram()
{
	Program* p = program->cloneQueryAndMetastatements(createdProgram);
	createdProgram = NULL;
	return p;
}

const Program* LogicProgramProblem::getBaseProgram()
{
	return program;
}

unsigned int LogicProgramProblem::getWidth() const
{
	return width;
}
