#include "asp-domain.hpp"

// In this file, output behaviour of asp entities are defined. I.e. the "write" function of each asp entity is defined.

using namespace std;
using namespace domain;
  
void TermTuple::write(ostream &stream) const
{
  for(vector<Term *>::iterator it = terms->begin(); it!=terms->end(); ++it)
  {
    (*it)->write(stream);
    if (it + 1 != terms->end())
      stream << ",";
  }
}

void TermTuples::write(ostream &stream) const
{
  for(vector<TermTuple *>::iterator it = tuples->begin(); it!=tuples->end(); ++it)
  {
    (*it)->write(stream);
    if (it + 1 != tuples->end())
      stream << ";";
  }
}

void UnaryTerm::write(ostream &stream) const
{
  switch(type)
  {
    case NEGATION:
      stream << "-";
      break;
    case ABSOLUTE:
      stream << "|";
      break;
    case COMPLEMENT:
      stream << "~";
      break;
    default:
      break;
  }
  base->write(stream);
  if (type == ABSOLUTE)
    stream << "|";
}

void BracketedTerm::write(ostream &stream) const
{
  stream << "(";
  base->write(stream);
  stream << ")";
}

void BinaryTerm::write(ostream &stream) const
{
  left->write(stream);
  stream << *binopType;
  right->write(stream);
}

void PoolingTerm::write(ostream &stream) const
{
  if (terms->size() == 1)
  { // actually no pooling; this term has been split
    terms->front()->write(stream);
    return;
  }
  stream << "(";
  for(vector<Term *>::iterator it = terms->begin(); it!=terms->end(); ++it)
  {
    (*it)->write(stream);
    if (it + 1 != terms->end())
      stream << ";";
  }
  stream << ")";
}

void Constant::write(ostream &stream) const
{
  stream << *printable;
}

void Variable::write(ostream &stream) const
{
  stream << *name;
}

void AnonymousVariable::write(ostream &stream) const
{
  stream << "_";
}

void Function::write(ostream &stream) const
{
  stream << *name;
  stream << "(";
  tuples->write(stream);
  stream << ")";
}

void ClassicalLiteral::write(ostream &stream) const
{
  if (nafNegated) stream << "not ";
  if (strongNegated) stream << "-";
  stream << *name;
  if (tuples)
  {
    stream << "(";
    tuples->write(stream);
    stream << ")";
  }
}

void Program::write(ostream &stream) const
{
  for(vector<Statement *>::iterator it = statements->begin(); it!=statements->end(); ++it)
  {
    (*it)->write(stream);
    stream << endl;
  }
  if (query) {
    query->write(stream);
    stream << "?";
    stream << endl;
  }
  for(vector<string *>::iterator it = metastatements->begin(); it!=metastatements->end(); ++it)
  {
    stream << **it << endl;
  }
}

void WeightAtLevel::write(ostream &stream) const
{
  weight->write(stream);
  if(level)
  {
    stream << "@";
    level->write(stream);
  }
  if(tuple)
  {
    stream << ",";
    tuple->write(stream);
  }
}

void OptimizeElement::write(ostream &stream) const
{
  weightAtLevel->write(stream);
  if(!literals->empty())
  {
    stream << ":";
    literals->write(stream);
  }
}

void OptimizeStatement::write(ostream &stream) const
{
  stream << *optimizeType;
  stream << "{";
    for(vector<OptimizeElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
    {
      (*it)->write(stream);
      if (it + 1 != elements->end())
	stream << ";";
    }
  stream << "}.";
}

void Body::write(ostream &stream) const
{
  for(vector<BodyElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
  {
    (*it)->write(stream);
    if (it + 1 != elements->end())
      stream << ", ";
  }
}

void WeakConstraint::write(ostream &stream) const
{
  stream << ":~ ";
  body->write(stream);
  stream << ". [";
  weightAtLevel->write(stream);
  stream << "]";
}

void Rule::write(ostream &stream) const
{
  head->write(stream);
  if (!body->empty())
  {
    if (!head->empty())
      stream << " ";
    stream << ":- ";
    body->write(stream);
  }
  if (head->empty() && body->empty())
    stream << ":-"; // totally empty rules would otherwise just be ".", which gringo does not accept.
  stream << ".";
}

void Disjunction::write(ostream &stream) const
{
  for(vector<ClassicalLiteral *>::iterator it = literals->begin(); it!=literals->end(); ++it)
  {
    (*it)->write(stream);
    if (it + 1 != literals->end())
      stream << " | ";
  }
}

void Guard::write(ostream &stream) const
{
  if(left)
  {
    term->write(stream);
    stream << *binop;
  }
  else // right
  {
    stream << *binop;
    term->write(stream);
  }
}

void ChoiceElement::write(ostream &stream) const
{
  literal->write(stream);
  if(!dependencyLiterals->empty())
  {
    stream << " : ";
    dependencyLiterals->write(stream);
  }
}

void Choice::write(ostream &stream) const
{
  if(guard_l)
    guard_l->write(stream);
  stream << "{";
  for(vector<ChoiceElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
  {
    (*it)->write(stream);
    if (it + 1 != elements->end())
      stream << "; ";
  }
  stream << "}";
  if(guard_r)
    guard_r->write(stream);
}

void AggregateElement::write(ostream &stream) const
{
  if(termTuple)
    termTuple->write(stream);
  if (!dependencyLiterals->empty())
  {
    stream << " : ";
    dependencyLiterals->write(stream);
  }
}

void Aggregate::write(ostream &stream) const
{
  if(nafNegated)
    stream << "not ";
  if(guard_l)
    guard_l->write(stream);
  stream << *function;
  stream << "{";
  for(vector<AggregateElement *>::iterator it = elements->begin(); it!=elements->end(); ++it)
  {
    (*it)->write(stream);
    if (it + 1 != elements->end())
      stream << "; ";
  }
  stream << "}";
  if(guard_r)
    guard_r->write(stream);
}

void BuiltinAtom::write(ostream &stream) const
{
  t1->write(stream);
  stream << *binop;
  t2->write(stream);
}

void LiteralBody::write(ostream &stream) const
{
  for(vector<NafLiteral *>::iterator it = literals->begin(); it!=literals->end(); ++it)
  {
    (*it)->write(stream);
    if (it + 1 != literals->end())
      stream << ", ";
  }
}
